import abc
from typing import TYPE_CHECKING, Optional, Type, Protocol


if TYPE_CHECKING:
    from .record import BaseRecord


def encode_table_id(table_id: str) -> str:
    import urllib.parse
    return urllib.parse.quote(table_id)


class _BaseAndTableProtocol(Protocol, metaclass=abc.ABCMeta):
    _API_ROOT_URL = 'https://api.airtable.com/v0'

    @property
    @abc.abstractmethod
    def base_id(self) -> Optional[str]:
        ...

    @property
    @abc.abstractmethod
    def table_id(self) -> Optional[str]:
        ...

    def build_url(self, record_id: Optional[str] = None) -> str:
        self._validate_base_table_ids()

        url = '%s/%s/%s' % (self._API_ROOT_URL, self.base_id, encode_table_id(self.table_id))
        if record_id is not None:
            url += '/%s' % record_id
        return url

    def ensure_base_and_table_match(self, other: '_BaseAndTableProtocol'):
        if self.base_id is not None and other.base_id is not None and self.base_id != other.base_id:
            raise ValueError('Base IDs do not match')
        if self.table_id is not None and other.table_id is not None and self.table_id != other.table_id:
            raise ValueError('Table IDs do not match')

    def _validate_base_table_ids(self) -> None:
        if self.base_id is None:
            raise ValueError('Base ID is not set')
        if self.table_id is None:
            raise ValueError('Table ID is not set')


class _BaseAndTableSettableProtocol(_BaseAndTableProtocol, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def set_base_id(self, base_id: str) -> '_BaseAndTableSettableMixin':
        ...

    @abc.abstractmethod
    def set_table_id(self, table_id: str) -> '_BaseAndTableSettableMixin':
        ...


class _BaseAndTableSettableMixin(_BaseAndTableSettableProtocol):
    def set_base_id(self, base_id: str) -> '_BaseAndTableSettableMixin':
        self._base_id = base_id
        return self

    def set_table_id(self, table_id: str) -> '_BaseAndTableSettableMixin':
        self._table_id = table_id
        return self

    @property
    def base_id(self) -> Optional[str]:
        return self._base_id

    @property
    def table_id(self) -> Optional[str]:
        return self._table_id

    _base_id: Optional[str]
    _table_id: Optional[str]

    def __init__(self,
                 record_cls: Optional[Type['BaseRecord']] = None,
                 base_id: Optional[str] = None,
                 table_id: Optional[str] = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._base_id = base_id
        if self._base_id is None and record_cls is not None:
            self._base_id = record_cls.get_class_base_id()
        self._table_id = table_id
        if self._table_id is None and record_cls is not None:
            self._table_id = record_cls.get_class_table_id()


class BaseAndTable(_BaseAndTableProtocol):
    _base_id: Optional[str]
    _table_id: Optional[str]

    def __init__(self, base_id: Optional[str] = None, table_id: Optional[str] = None):
        self._base_id = base_id
        self._table_id = table_id

    @property
    def base_id(self) -> Optional[str]:
        return self._base_id

    @property
    def table_id(self) -> Optional[str]:
        return self._table_id

    def __repr__(self):
        return '%s(base_id=%r, table_id=%r)' % (self.__class__.__name__, self._base_id, self._table_id)


__all__ = ['_BaseAndTableProtocol', '_BaseAndTableSettableProtocol', '_BaseAndTableSettableMixin', 'BaseAndTable']



from typing import TYPE_CHECKING, Type, Iterator, Optional, Union, Dict, Iterable
from ..exceptions import RequestError


try:
    import simplejson as json
except ImportError:
    import json


if TYPE_CHECKING:
    from pyrtable._baseandtable import _BaseAndTableProtocol
    from pyrtable.filters.base import BaseFilter
    from pyrtable.query import RecordQuery
    from pyrtable.record import BaseRecord


# noinspection PyMethodMayBeStatic
class BaseContext:
    def fetch_single(self, *,
                     record_cls: Type['BaseRecord'],
                     record_id: str,
                     base_and_table: '_BaseAndTableProtocol') \
            -> 'BaseRecord':
        import requests
        from pyrtable.connectionmanager import get_connection_manager

        headers = record_cls.get_request_headers(base_id=base_and_table.base_id)
        url = base_and_table.build_url(record_id=record_id)

        with get_connection_manager():
            response = requests.get(url, headers=headers)
            if 400 <= response.status_code < 500:
                error = response.json().get('error', {})

                if error == 'NOT_FOUND':
                    raise KeyError(record_id)

                error_message = error.get('message', '')
                error_type = error.get('type', '')

                raise RequestError(message=error_message, type=error_type)

        record_data = response.json()
        record = record_cls(_base_id=base_and_table.base_id, _table_id=base_and_table.table_id)
        record.consume_airtable_data(record_data)
        return record

    def fetch_many(self, *,
                   record_cls: Type['BaseRecord'],
                   base_and_table: '_BaseAndTableProtocol',
                   record_filter: Optional['BaseFilter'] = None) \
            -> Iterator['BaseRecord']:
        import urllib.parse
        import requests
        from pyrtable.connectionmanager import get_connection_manager

        headers = record_cls.get_request_headers(base_id=base_and_table.base_id)
        url = base_and_table.build_url()
        parsed_url = urllib.parse.urlparse(url)
        url_query_params = urllib.parse.parse_qsl(parsed_url.query, keep_blank_values=True)

        if record_filter:
            filter_by_formula = record_filter.build_formula(record_cls)
            if filter_by_formula:
                url_query_params.append(('filterByFormula', filter_by_formula))

        column_names = record_cls.get_column_names()
        url_query_params.extend(('fields[]', column_name) for column_name in column_names)

        # noinspection PyProtectedMember
        parsed_url = parsed_url._replace(query=urllib.parse.urlencode(url_query_params))

        while True:
            with get_connection_manager():
                response = requests.get(parsed_url.geturl(), headers=headers)
                if 400 <= response.status_code < 500:
                    error = response.json().get('error', {})
                    error_message = error.get('message', '')
                    error_type = error.get('type', '')

                    raise RequestError(message=error_message, type=error_type)

            response_json = response.json()
            for record_data in response_json.get('records', []):
                record = record_cls(_base_id=base_and_table.base_id, _table_id=base_and_table.table_id)
                record.consume_airtable_data(record_data)
                yield record

            offset = response_json.get('offset')
            if offset is None:
                break

            url_query_params = list(filter(lambda pair: pair[0] != 'offset', url_query_params))
            url_query_params.append(('offset', offset))

            # noinspection PyProtectedMember
            parsed_url = parsed_url._replace(query=urllib.parse.urlencode(url_query_params))

    def _create(self,
                record_cls: Type['BaseRecord'],
                record: 'BaseRecord') -> None:
        import requests
        from pyrtable.connectionmanager import get_connection_manager

        url = record.build_url()
        headers = record_cls.get_request_headers({
            'Content-Type': 'application/json',
        }, base_id=record.base_id)
        data = {'fields': record.encode_to_airtable()}

        with get_connection_manager():
            response = requests.post(url, headers=headers, data=json.dumps(data))
            if 400 <= response.status_code < 500:
                error = response.json().get('error', {})
                error_message = error.get('message', '')
                error_type = error.get('type', '')

                raise RequestError(message=error_message, type=error_type)

        record.consume_airtable_data(response.json())

    def _update(self,
                record_cls: Type['BaseRecord'],
                record: 'BaseRecord') -> None:
        import requests
        from pyrtable.connectionmanager import get_connection_manager

        dirty_fields = record.encode_to_airtable()
        if not dirty_fields:
            return

        url = record.build_url(record_id=record.id)
        headers = record_cls.get_request_headers({
            'Content-Type': 'application/json',
        }, base_id=record.base_id)
        data = {'fields': dirty_fields}

        with get_connection_manager():
            response = requests.patch(url, headers=headers, data=json.dumps(data))
            if 400 <= response.status_code < 500:
                error = response.json().get('error', {})
                error_message = error.get('message', '')
                error_type = error.get('type', '')

                raise RequestError(message=error_message, type=error_type)

        # noinspection PyProtectedMember
        record._clear_dirty_fields()

    def save(self,
             record_cls: Type['BaseRecord'],
             record: 'BaseRecord') -> None:
        """
        Save the record to Airtable.
        """
        if record.id is None:
            self._create(record_cls, record)
        else:
            self._update(record_cls, record)

    def delete_id(self, *,
                  record_cls: Type['BaseRecord'],
                  record_id: str,
                  base_and_table: '_BaseAndTableProtocol') -> None:
        import requests
        from pyrtable.connectionmanager import get_connection_manager

        url = base_and_table.build_url(record_id=record_id)
        headers = record_cls.get_request_headers(base_id=base_and_table.base_id)

        with get_connection_manager():
            response = requests.delete(url, headers=headers)
            if 400 <= response.status_code < 500:
                error = response.json().get('error', {})
                error_message = error.get('message', '')
                error_type = error.get('type', '')

                raise RequestError(message=error_message, type=error_type)

    def delete(self,
               record_cls: Type['BaseRecord'],
               record: 'BaseRecord') -> None:
        if record.id is None:
            return

        self.delete_id(record_cls=record_cls, record_id=record.id, base_and_table=record)

        record._id = None
        record._created_timestamp = None
        # delete() does not clear data, but anything that is semantically different from None
        # becomes instantly dirty. That's why we are assigning None to all original fields values.
        for attr_name, field in record.iter_fields():
            # noinspection PyProtectedMember
            record._orig_fields_values[attr_name] = field.decode_from_airtable(None, base_and_table=record)


class SimpleCachingContext(BaseContext):
    @staticmethod
    def _build_key(record_cls: Type['BaseRecord'], base_and_table: '_BaseAndTableProtocol', record_id: str) -> str:
        # noinspection PyProtectedMember
        base_and_table._validate_base_table_ids()
        return '%s:%s:%s' % (record_cls.__name__, base_and_table.base_id, record_id)

    _cache: Dict[str, 'BaseRecord']

    def __init__(self,
                 allow_classes: Optional[Iterable[Type['BaseRecord']]] = None,
                 exclude_classes: Optional[Iterable[Type['BaseRecord']]] = None):
        import threading

        self._cache = {}
        self._cache_lock = threading.Lock()
        self._allow_classes = set(allow_classes) if allow_classes is not None else None
        self._exclude_classes = set(exclude_classes) if exclude_classes is not None else None

    def _is_cached_class(self, record_cls: Type['BaseRecord']):
        if self._exclude_classes is not None and record_cls in self._exclude_classes:
            return False
        if self._allow_classes is not None:
            return record_cls in self._allow_classes
        return True

    def pre_cache(self, *args: Union['BaseRecord', 'RecordQuery', Type['BaseRecord']]):
        import inspect
        from pyrtable.query import RecordQuery
        from pyrtable.record import BaseRecord

        for arg in args:
            if isinstance(arg, BaseRecord) and arg.id is not None:
                with self._cache_lock:
                    self._cache[self._build_key(arg.__class__, arg, arg.id)] = arg
            elif isinstance(arg, RecordQuery):
                # This will fetch and cache records
                _ = list(arg)
            elif inspect.isclass(arg) and issubclass(arg, BaseRecord):
                # This will fetch and cache records
                _ = list(arg.objects.all())
            else:
                raise ValueError(arg)

    def fetch_single(self, *,
                     record_cls: Type['BaseRecord'],
                     record_id: str,
                     base_and_table: '_BaseAndTableProtocol') -> 'BaseRecord':
        if not self._is_cached_class(record_cls):
            return super(SimpleCachingContext, self).fetch_single(
                record_cls=record_cls, record_id=record_id, base_and_table=base_and_table)

        key = self._build_key(record_cls, base_and_table, record_id)
        with self._cache_lock:
            record = self._cache.get(key)
        if record is not None:
            return record

        record = super(SimpleCachingContext, self).fetch_single(
            record_cls=record_cls, record_id=record_id, base_and_table=base_and_table)
        with self._cache_lock:
            self._cache[key] = record
        return record

    def fetch_many(self, *,
                   record_cls: Type['BaseRecord'],
                   base_and_table: '_BaseAndTableProtocol',
                   record_filter: Optional['BaseFilter'] = None) -> Iterator['BaseRecord']:
        for record in super(SimpleCachingContext, self).fetch_many(
                record_cls=record_cls, base_and_table=base_and_table, record_filter=record_filter):
            if self._is_cached_class(record_cls):
                with self._cache_lock:
                    self._cache[self._build_key(record_cls, record, record.id)] = record
            yield record

    def save(self, record_cls: Type['BaseRecord'], record: 'BaseRecord') -> None:
        super(SimpleCachingContext, self).save(record_cls, record)
        if self._is_cached_class(record_cls):
            with self._cache_lock:
                self._cache[self._build_key(record_cls, record, record.id)] = record

    def delete_id(self, *,
                  record_cls: Type['BaseRecord'],
                  record_id: str,
                  base_and_table: '_BaseAndTableProtocol') -> None:
        if self._is_cached_class(record_cls):
            with self._cache_lock:
                self._cache.pop(self._build_key(record_cls, base_and_table, record_id), None)

        super(SimpleCachingContext, self).delete_id(
            record_cls=record_cls, record_id=record_id, base_and_table=base_and_table)


__all__ = ['BaseContext', 'SimpleCachingContext']



from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, Optional, List

from .base import BaseField


if TYPE_CHECKING:
    from pyrtable._baseandtable import _BaseAndTableProtocol


@dataclass
class _AttachmentThumbnail:
    url: str
    width: int
    height: int

    def download(self) -> bytes:
        import requests

        response = requests.get(self.url)
        return response.content


@dataclass
class _AttachmentThumbnailSet:
    small: Optional[_AttachmentThumbnail] = None
    large: Optional[_AttachmentThumbnail] = None
    full: Optional[_AttachmentThumbnail] = None

    def __post_init__(self):
        if self.small is not None:
            self.small = _AttachmentThumbnail(**self.small)
        if self.large is not None:
            self.large = _AttachmentThumbnail(**self.large)
        if self.full is not None:
            self.full = _AttachmentThumbnail(**self.full)


@dataclass
class Attachment:
    id: str
    url: str
    filename: str
    size: int
    type: str
    width: Optional[int] = None
    height: Optional[int] = None
    thumbnails: Optional[_AttachmentThumbnailSet] = None

    def __post_init__(self):
        if self.thumbnails is not None:
            self.thumbnails = _AttachmentThumbnailSet(**self.thumbnails)

    def download(self) -> bytes:
        import requests

        response = requests.get(self.url)
        return response.content

    def download_to(self, path):
        contents = self.download()
        with open(path, 'wb') as fd:
            fd.write(contents)


class AttachmentField(BaseField):
    def decode_from_airtable(self, value: Optional[List[Dict[str, Any]]], base_and_table: '_BaseAndTableProtocol')\
            -> List[Attachment]:
        if not value:
            return []
        return [Attachment(**item_data) for item_data in value]

    def encode_to_airtable(self, value):
        raise NotImplementedError('AttachmentField is only implemented as a read-only field')

    def __init__(self, *args, **kwargs):
        if not kwargs.get('read_only', False):
            raise AttributeError('AttachmentField is only implemented as a read-only field')
        super().__init__(*args, **kwargs)


__all__ = ['Attachment', 'AttachmentField']



from __future__ import annotations

from abc import ABC
from enum import Enum
from typing import TYPE_CHECKING, List, Tuple, Any, Union, Type, Optional, Iterable

from .base import BaseField
from .valueset import ValueSet, validator_from_enum


if TYPE_CHECKING:
    from pyrtable._baseandtable import _BaseAndTableProtocol


class _AbstractSelectionField(BaseField, ABC):
    _enum_class: Optional[Type[Enum]] = None

    def __init__(self, column_name: str, choices: Optional[Union[Type[Enum], List[Tuple[Any, Any]]]] = None, **kwargs):
        if choices is not None:
            import inspect

            if inspect.isclass(choices) and issubclass(choices, Enum):
                self._enum_class = choices

                choices = list((value, value.value) for value in choices)

            self._raw_to_value = {b: a for a, b in choices}
            self._value_to_raw = {a: b for a, b in choices}

        else:
            self._raw_to_value = {}
            self._value_to_raw = {}

        super().__init__(column_name, **kwargs)


class SingleSelectionField(_AbstractSelectionField):
    def decode_from_airtable(self, value: Optional[str], base_and_table: '_BaseAndTableProtocol') -> Optional[Any]:
        return self._raw_to_value.get(value, value)

    def encode_to_airtable(self, value: Optional[Any]) -> Optional[str]:
        return self._value_to_raw.get(value, value)


class MultipleSelectionField(_AbstractSelectionField):
    def _create_value_set(self):
        if self._enum_class is not None:
            validator = validator_from_enum(self._enum_class)
        else:
            validator = None
        return ValueSet(validator=validator)

    def clone_value(self, value: ValueSet) -> ValueSet:
        return ValueSet(value)

    def decode_from_airtable(self, value: Optional[List[str]], base_and_table: '_BaseAndTableProtocol') -> ValueSet:
        if value is not None:
            value = [self._raw_to_value.get(item, item) for item in value]
        return self.validate(value, base_and_table=base_and_table)

    def encode_to_airtable(self, value: Optional[ValueSet]) -> Optional[List[str]]:
        if not value:
            return None
        return [self._value_to_raw.get(item, item) for item in value]

    def validate(self, value: Optional[Iterable[Any]], base_and_table: '_BaseAndTableProtocol') -> ValueSet:
        value_set = self._create_value_set()
        if value is not None:
            value_set.set(value)
        return value_set


__all__ = ['SingleSelectionField', 'MultipleSelectionField']



from typing import Type, TYPE_CHECKING


if TYPE_CHECKING:
    from pyrtable.fields import BaseField
    from pyrtable.record import BaseRecord


class BaseFilter:
    @staticmethod
    def get_field_object(record_class: Type['BaseRecord'], attr_name: str) -> 'BaseField':
        try:
            return next(field for field_attr_name, field in record_class.iter_fields()
                        if field_attr_name == attr_name)
        except StopIteration:
            raise AttributeError("%r has no attribute %r" % (record_class, attr_name))

    @staticmethod
    def get_column_name(record_class: Type['BaseRecord'], attr_name: str) -> str:
        return BaseFilter.get_field_object(record_class=record_class, attr_name=attr_name).column_name

    def __and__(self, other):
        from .raw import AndFilter

        if not isinstance(other, BaseFilter):
            return NotImplemented

        return AndFilter(self, other)

    def __or__(self, other):
        from .raw import OrFilter

        if not isinstance(other, BaseFilter):
            return NotImplemented

        return OrFilter(self, other)

    def __invert__(self):
        from .raw import NotFilter

        if isinstance(self, NotFilter):
            return self.filter
        return NotFilter(self)

    def build_formula(self, record_class: Type['BaseRecord']) -> str:
        raise NotImplementedError()


__all__ = ['BaseFilter']



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Any, Callable, Type


if TYPE_CHECKING:
    from pyrtable._baseandtable import _BaseAndTableProtocol
    from pyrtable.record import BaseRecord


class BaseField:
    attr_name: Optional[str] = None
    default_value: Optional[Any] = None
    normalize: Optional[Callable[[Any], Any]] = None
    normalize_from_attr_name: Optional[str] = None
    skip_normalization_if_filled: bool
    _record: Optional[BaseRecord]

    @classmethod
    def _install_extra_properties(cls, record_cls: Type['BaseRecord'], attr_name: str):
        pass

    def __init__(self, column_name: str, default=None, read_only: Optional[bool] = None,
                 normalize=None, normalize_from=None,
                 skip_normalization_if_filled=False):
        if read_only and default is not None:
            raise ValueError('Cannot set a default value for a read-only field')

        self._column_name = column_name
        self._read_only = read_only
        self.default_value = default
        self.normalize = normalize
        self.normalize_from_attr_name = normalize_from
        self.skip_normalization_if_filled = skip_normalization_if_filled

    @property
    def column_name(self) -> Optional[str]:
        return self._column_name

    @property
    def read_only(self):
        if self._read_only is not None:
            return self._read_only
        # noinspection PyProtectedMember
        record_read_only = self._record._get_meta_attr('read_only', None)
        if record_read_only is not None:
            return record_read_only
        return False

    def __get__(self, instance: BaseRecord, owner):
        # noinspection PyProtectedMember
        return instance._fields_values.get(self.attr_name)

    def __set__(self, instance: BaseRecord, value):
        if self.read_only and self._record.id is not None:
            raise AttributeError('%s: This field is read-only' % self.attr_name)
        # noinspection PyProtectedMember
        instance._fields_values[self.attr_name] = self.validate(value, base_and_table=instance)

    @classmethod
    def is_same_value(cls, lhs, rhs) -> bool:
        return lhs == rhs

    def validate(self, value: Any, base_and_table: '_BaseAndTableProtocol') -> Any:
        """Validate the value before storing it.

        Validation means changing the value to appropriate form (e.g., making sure it's converted to a specific class or
        stripping spaces) and/or raising an exception if the value cannot be accepted.

        :param value: The value that's being stored.
        :param base_and_table: The base-and-table info holder. Usually the record itself; in case of linked records,
        this is the base-and-table that holds the link (not the linked one).
        :return: The accepted value for storage.
        :raise: `ValueError` if the value value cannot be accepted.
        """
        return value

    def decode_from_airtable(self, value, base_and_table: '_BaseAndTableProtocol'):
        """
        Decode a field value retrieved through Airtable JSON API into the corresponding Python value.

        :param value: The field value retrieved from the server.
        :param base_and_table: The base-and-table info holder. Usually the record itself; in case of linked records,
        this is the base-and-table that holds the link (not the linked one).
        :return: The corresponding Python value.
        """
        raise NotImplementedError()

    def encode_to_airtable(self, value):
        """
        Encode a native Python value into a field value ready to send to Airtable JSON API.

        :param value: The Python value.
        :return: The corresponding field value ready to send to the server.
        """
        raise NotImplementedError()

    def clone_value(self, value: Any) -> Any:
        """
        Clone (create an independent copy of) a value so it can be compared later for changes.

        :param value: Value to be cloned.
        :return: Cloned value.
        """
        return value

    def __repr__(self):
        if self.attr_name is None:
            attr_name = '(unbound field)'
        else:
            attr_name = self.attr_name

        return '<%s: %s>' % (self.__class__.__name__, attr_name)



import collections.abc
from typing import TYPE_CHECKING, Generic, Iterable, Iterator, TypeVar, Type, Optional

from ._baseandtable import _BaseAndTableSettableProtocol, BaseAndTable


if TYPE_CHECKING:
    from .filters.base import BaseFilter
    from .record import BaseRecord


RT = TypeVar('RT', bound='BaseRecord')
QT = TypeVar('QT', bound='RecordQuery')


class RecordQuery(BaseAndTable, Generic[RT, QT], Iterable[RT],
                  collections.abc.Iterable, _BaseAndTableSettableProtocol):
    """
    A (potentially under construction) query for records in a table. Also represents the starting point for queries
    to be made over a :class:`BaseRecord` derived class, exposed through the `objects` class attribute.
    """
    def set_base_id(self, base_id: str) -> 'QT':
        """
        Change the query's base ID.

        :return: The resulting query.
        """
        result = self._shallow_copy()
        result._base_id = base_id
        return result

    def set_table_id(self, table_id: str) -> 'QT':
        """
        Change the query's table ID.

        :return: The resulting query.
        """
        result = self._shallow_copy()
        result._table_id = table_id
        return result

    _record_class: Type['BaseRecord']
    _initialised = False
    _is_empty_query = False

    def __init__(self, record_class: Type['BaseRecord'], flt: Optional['BaseFilter'] = None):
        super().__init__(base_id=record_class.get_class_base_id(), table_id=record_class.get_class_table_id())

        self._record_class = record_class
        self._filter = flt

    def _shallow_copy(self) -> QT:
        import copy
        result = copy.copy(self)
        return result

    def _shallow_copy_and_initialise(self) -> QT:
        result = self._shallow_copy()
        result._initialised = True
        return result

    def all(self) -> QT:
        """
        Return a query for all records, given that they are not filtered out by other criteria.

        :return: The resulting query.
        """
        if self._initialised:
            return self
        return self._shallow_copy_and_initialise()

    def none(self) -> QT:
        """
        Return an empty query. No server communication is needed to execute this query.

        :return: The resulting query.
        """
        if self._is_empty_query:
            return self
        result = self._shallow_copy_and_initialise()
        result._is_empty_query = True
        return result

    def filter(self, *args, **kwargs) -> QT:
        """
        Return a query that will respect the criteria given as arguments.

        :return: The resulting query.
        """
        if not args and not kwargs:
            return self.all()

        from .filters import Q

        result = self._shallow_copy_and_initialise()

        if result._filter is None:
            result._filter = Q(*args, **kwargs)
        elif isinstance(result._filter, Q):
            result._filter.extend(*args, **kwargs)
        else:
            result._filter = Q(result._filter, *args, **kwargs)

        return result

    def get(self, record_id: str) -> RT:
        """
        Return a single record with given identifier, if it exists in the table.

        :return: The matching record.
        :raises KeyError: if no record matches the given record ID.
        :raises ValueError: if filters were applied (currently this is not supported).
        """
        from pyrtable.context import get_default_context

        # Trivial implementation for Record.objects.none().get(...)
        if self._is_empty_query:
            raise KeyError(record_id)

        if self._filter is not None:
            raise ValueError('Currently get() is not compatible with filters applied')

        return get_default_context().fetch_single(
            record_cls=self._record_class, record_id=record_id, base_and_table=self)

    def __iter__(self) -> Iterator[RT]:
        if not self._initialised:
            raise ValueError('Query is not initialised. Use .all(), .filter() or .none() to initialise it.')

        if self._is_empty_query:
            return

        from pyrtable.context import get_default_context
        yield from get_default_context().fetch_many(
            record_cls=self._record_class, base_and_table=self, record_filter=self._filter)


__all__ = ['RecordQuery']



import datetime
import threading
import time
from typing import Optional


class _ConnectionManager:
    def __init__(self, min_interval_seconds: float):
        self._lock = threading.Lock()
        self._last_timestamp = datetime.datetime.min
        self._min_interval_seconds = min_interval_seconds

    def __enter__(self):
        with self._lock:
            delta_seconds = (datetime.datetime.now() - self._last_timestamp).total_seconds()
            if delta_seconds < self._min_interval_seconds:
                wait_time = self._min_interval_seconds - delta_seconds
                time.sleep(wait_time)

            self._last_timestamp = datetime.datetime.now()

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


_connection_manager: Optional[_ConnectionManager] = None
_connection_manager_lock = threading.Lock()


def get_connection_manager() -> _ConnectionManager:
    global _connection_manager

    if _connection_manager is None:
        with _connection_manager_lock:
            if _connection_manager is None:
                _connection_manager = _ConnectionManager(min_interval_seconds=0.2)

    return _connection_manager



import re
from typing import TYPE_CHECKING, Type, Dict, Tuple

from .base import BaseFilter
from .raw import *


if TYPE_CHECKING:
    from pyrtable.record import BaseRecord


ALL_FILTERS: Dict[str, Type['BaseFilter']] = {
    '': EqualsFilter,
    'ne': NotEqualsFilter,
    'gt': GreaterThanFilter,
    'lt': LessThanFilter,
    'gte': GreaterThanOrEqualsFilter,
    'lte': LessThanOrEqualsFilter,
    'ge': GreaterThanOrEqualsFilter,
    'le': LessThanOrEqualsFilter,
    'empty': IsEmptyFilter,
    'contains': ContainsFilter,
    'excludes': DoesNotContainFilter,
}


def get_filter(arg_name: str) -> Tuple[Type['BaseFilter'], str]:
    m = re.fullmatch(r'(?P<attr_name>.+)__(?P<operation>[a-z]+)', arg_name)
    if not m:
        return EqualsFilter, arg_name

    operation = m.group('operation')
    if operation not in ALL_FILTERS:
        raise ValueError('Invalid operation "%s"' % operation)

    return ALL_FILTERS[operation], m.group('attr_name')


class Q(BaseFilter):
    def build_formula(self, record_class: Type['BaseRecord']) -> str:
        return self._filter.build_formula(record_class)

    def __init__(self, *args, **kwargs):
        self._filter = AndFilter()
        self.extend(*args, **kwargs)

    def extend(self, *args, **kwargs):
        for key, value in kwargs.items():
            filter_cls, attr_name = get_filter(key)
            self._filter.filters.append(filter_cls(attr_name, value))

        for arg in args:
            if not isinstance(arg, BaseFilter):
                raise ValueError(arg)
            self._filter.filters.append(arg)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self._filter)


__all__ = ['Q']




# -*- coding: utf-8 -*-

# DO NOT EDIT THIS FILE!
# This file has been autogenerated by dephell <3
# https://github.com/dephell/dephell

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


import os.path

readme = ''
here = os.path.abspath(os.path.dirname(__file__))
readme_path = os.path.join(here, 'README.rst')
if os.path.exists(readme_path):
    with open(readme_path, 'rb') as stream:
        readme = stream.read().decode('utf8')


setup(
    long_description=readme,
    name='pyrtable',
    version='0.8.2',
    description='Django-inspired library to interface with Airtable',
    python_requires='==3.*,>=3.8.0',
    project_urls={"documentation": "https://pyrtable.readthedocs.io/", "repository": "https://github.com/vilarneto/pyrtable"},
    author='Vilar da Camara Neto',
    author_email='vilarneto@gmail.com',
    license='MIT',
    classifiers=['Development Status :: 4 - Beta', 'Environment :: Console', 'Intended Audience :: Developers', 'License :: OSI Approved :: MIT License', 'Natural Language :: English', 'Operating System :: OS Independent', 'Programming Language :: Python', 'Programming Language :: Python :: 3.8', 'Programming Language :: Python :: 3.9', 'Programming Language :: Python :: 3.10', 'Programming Language :: Python :: Implementation :: CPython', 'Topic :: Software Development'],
    packages=['pyrtable', 'pyrtable.context', 'pyrtable.fields', 'pyrtable.filters'],
    package_dir={"": "src"},
    package_data={},
    install_requires=['deprecated', 'pytz', 'pyyaml==5.*,>=5.1.0', 'requests==2.*,>=2.22.0', 'simplejson==3.*,>=3.16.0'],
)



from __future__ import annotations

import collections.abc
from enum import Enum
from typing import TypeVar, MutableSet, Iterator, Callable, Optional, Union, Any, Type, Generic, Iterable


T = TypeVar('T')
T_co = TypeVar('T_co', covariant=True)


class ValueSet(collections.abc.MutableSet, collections.abc.Iterable, Generic[T]):
    _items: MutableSet[T]
    _validator: Callable[[T], T]

    def __init__(self, other: Optional[ValueSet] = None, *, validator: Optional[Callable[[Any], T]] = None):
        if other is not None:
            self._items = set(other._items)
            self._validator = other._validator
        else:
            self._items = set()

        if validator is not None:
            self._validator = validator

        if self._validator is None:
            def validator(value: T) -> T:
                return value

            self._validator = validator

    def add(self, x: T) -> None:
        self._items.add(self._validator(x))

    def discard(self, x: T) -> None:
        self._items.discard(self._validator(x))

    def set(self, iterable: Iterable[T]) -> None:
        self._items = set(self._validator(value) for value in iterable)

    def __contains__(self, x: Any) -> bool:
        return self._items.__contains__(self._validator(x))

    def __len__(self) -> int:
        return self._items.__len__()

    def __iter__(self) -> Iterator[T_co]:
        return self._items.__iter__()

    def __iadd__(self, other: T) -> ValueSet[T]:
        self.add(other)
        return self

    def __isub__(self, other: T) -> ValueSet[T]:
        self.discard(other)
        return self

    def __repr__(self) -> str:
        return self._items.__repr__()


def validator_from_enum(enum_class: Type[Enum]) -> Callable[[Union[str, Enum]], Union[str, Enum]]:
    def validator(value: Union[str, Enum]) -> Union[str, Enum]:
        try:
            return enum_class(value)
        except ValueError:
            return value

    return validator


__all__ = ['ValueSet', 'validator_from_enum']



import re
from enum import Enum


def quote_column_name(column_name: str) -> str:
    return '{%s}' % re.sub(r'[}\\]', lambda ch: '\\' + ch.group(0), column_name)


def quote_value(value) -> str:
    import datetime

    if isinstance(value, str):
        return '"%s"' % re.sub(r'(["\\])', lambda ch: '\\' + ch.group(0), value)
    if isinstance(value, bool):
        return 'TRUE()' if value else 'FALSE()'
    if isinstance(value, (int, float)):
        return str(value)
    if isinstance(value, Enum):
        return quote_value(value.value)
    if isinstance(value, datetime.datetime):
        return format(value, '"%Y-%m-%dT%H:%M:%S.%fZ"')
    if isinstance(value, datetime.date):
        return format(value, '"%Y-%m-%d"')
    raise ValueError(value)


def airtable_filter_and(*inner_filters) -> str:
    inner_filters = [inner_filter for inner_filter in inner_filters if inner_filter]

    if len(inner_filters) == 0:
        return ''
    if len(inner_filters) == 1:
        return inner_filters[0]
    return 'AND(%s)' % ','.join(inner_filters)


def airtable_filter_or(*inner_filters) -> str:
    inner_filters = [inner_filter for inner_filter in inner_filters if inner_filter]

    if len(inner_filters) == 0:
        return ''
    if len(inner_filters) == 1:
        return inner_filters[0]
    return 'OR(%s)' % ','.join(inner_filters)


def airtable_filter_not(inner_filter) -> str:
    if not inner_filter:
        return 'FALSE()'
    return 'NOT(%s)' % inner_filter


def airtable_filter_equals(column_name, value) -> str:
    return '{column_name}={value}'.format(
        column_name=quote_column_name(column_name), value=quote_value(value))


def airtable_filter_not_equals(column_name, value) -> str:
    return '{column_name}!={value}'.format(
        column_name=quote_column_name(column_name), value=quote_value(value))


def airtable_filter_greater_than(column_name, value) -> str:
    return '{column_name}>{value}'.format(
        column_name=quote_column_name(column_name), value=quote_value(value))


def airtable_filter_less_than(column_name, value) -> str:
    return '{column_name}<{value}'.format(
        column_name=quote_column_name(column_name), value=quote_value(value))


def airtable_filter_greater_than_or_equals(column_name, value) -> str:
    return '{column_name}>={value}'.format(
        column_name=quote_column_name(column_name), value=quote_value(value))


def airtable_filter_less_than_or_equals(column_name, value) -> str:
    return '{column_name}<={value}'.format(
        column_name=quote_column_name(column_name), value=quote_value(value))


def airtable_filter_multiple_select_contains(column_name, value):
    return 'FIND(", "&{value}&", ",", "&{column_name}&", ")>0'.format(
        column_name=quote_column_name(column_name), value=quote_value(value))


def airtable_filter_multiple_select_does_not_contain(column_name, value):
    return 'FIND(", "&{value}&", ",", "&{column_name}&", ")=0'.format(
        column_name=quote_column_name(column_name), value=quote_value(value))


def airtable_filter_multiple_select_is_empty(column_name):
    return 'NOT({column_name})'.format(column_name=quote_column_name(column_name))


def airtable_filter_multiple_select_is_not_empty(column_name):
    return '{column_name}!=""'.format(column_name=quote_column_name(column_name))



from __future__ import annotations

import collections.abc
from abc import ABCMeta
from typing import TYPE_CHECKING, Optional, Union, Type, List, Iterator, Iterable, Any

from pyrtable._baseandtable import _BaseAndTableProtocol, BaseAndTable
from pyrtable.fields import BaseField


if TYPE_CHECKING:
    from pyrtable.record import BaseRecord


class BaseRecordLinkField(BaseField, metaclass=ABCMeta):
    _linked_class: Union[Type[BaseRecord], str]

    def __init__(self, column_name: str,
                 linked_class: Union[Type[BaseRecord], str],
                 *args, **kwargs):
        from pyrtable.record import BaseRecord

        def fetcher(linked_record_id: str, *, base_and_table: '_BaseAndTableProtocol') -> BaseRecord:
            nonlocal linked_class

            if isinstance(linked_class, str):
                import re

                module_name, class_name = re.fullmatch(r'(.*)\.(.+)', linked_class).groups()
                module = __import__(module_name, fromlist=[class_name])
                linked_class = getattr(module, class_name)

            query = linked_class.objects
            if base_and_table.base_id is not None:
                query = query.set_base_id(base_and_table.base_id)
            if base_and_table.table_id is not None:
                query = query.set_table_id(base_and_table.table_id)
            return query.get(record_id=linked_record_id)

        self._linked_class = linked_class
        self._fetcher = fetcher

        super().__init__(column_name, *args, **kwargs)

    def get_linked_class(self) -> Type[BaseRecord]:
        if isinstance(self._linked_class, str):
            import re

            module_name, class_name = re.fullmatch(r'(.*)\.(.+)', self._linked_class).groups()
            module = __import__(module_name, fromlist=[class_name])
            self._linked_class = getattr(module, class_name)

        return self._linked_class

    def get_linked_base_and_table(self) -> '_BaseAndTableProtocol':
        from pyrtable._baseandtable import BaseAndTable

        linked_class = self.get_linked_class()
        base_id = linked_class._get_meta_attr('base_id', None)
        table_id = linked_class._get_meta_attr('table_id', None)
        return BaseAndTable(base_id, table_id)


class _RecordLink(BaseAndTable):
    _id: Optional[str] = None
    _record: Optional[BaseRecord] = None
    _fetcher = None

    def __init__(self, *,
                 base_and_table: '_BaseAndTableProtocol',
                 other: Optional[_RecordLink] = None,
                 record_id: Optional[str] = None,
                 record: Optional[BaseRecord] = None,
                 fetcher=None):
        super().__init__(base_id=base_and_table.base_id, table_id=base_and_table.table_id)

        if other is not None:
            self._id = other._id
            self._record = other._record
            self._fetcher = other._fetcher

        if record_id is not None:
            self._id = record_id
        if record is not None:
            self._record = record
        if fetcher is not None:
            self._fetcher = fetcher

    @property
    def id(self) -> str:
        if self._record is not None and self._record.id is not None:
            return self._record.id
        if self._id is None:
            raise ValueError('Reference to unsaved or deleted record')
        return self._id

    @property
    def record(self) -> BaseRecord:
        if self._record is None:
            if self._fetcher is None:
                # @TODO Better error message
                raise ValueError("Don't know how to fetch a record - use `fetcher` or `linked_class` attributes")
            self._record = self._fetcher(self._id, base_and_table=self)

        return self._record

    @property
    def has_fetched_record(self):
        return self._record is not None

    def __eq__(self, other):
        if not isinstance(other, _RecordLink):
            return NotImplemented
        if self._id is not None and self._id == other._id:
            return True
        return self._record is not None and self._record is other._record


class _SingleRecordIdLinkProxyField:
    def __init__(self, record_attr_name: str):
        self._record_attr_name = record_attr_name

    def __get__(self, instance, owner: 'BaseField'):
        # noinspection PyProtectedMember
        value = instance._fields_values[self._record_attr_name]
        return value.id if value is not None else None

    def __set__(self, instance, value: Optional[str]):
        if value is None:
            # noinspection PyProtectedMember
            instance._fields_values[self._record_attr_name] = value
            return
        # @TODO Not really that; also, ignore if value does not change
        raise ValueError(value)


class SingleRecordLinkField(BaseRecordLinkField):
    @staticmethod
    def _get_id_attr_name(attr_name: str) -> str:
        return '%s_id' % attr_name

    @classmethod
    def _install_extra_properties(cls, record_cls: Type['BaseRecord'], attr_name: str):
        super()._install_extra_properties(record_cls, attr_name)
        setattr(record_cls, cls._get_id_attr_name(attr_name), _SingleRecordIdLinkProxyField(attr_name))

    def __get__(self, instance, owner: 'BaseRecord') -> Optional['BaseRecord']:
        value = super().__get__(instance, owner)
        return value.record if value is not None else None

    def validate(self, value: Optional[Union[_RecordLink, Iterable[Any]]], base_and_table: '_BaseAndTableProtocol') \
            -> Any:
        from pyrtable.record import BaseRecord

        if isinstance(value, _RecordLink):
            return value
        if isinstance(value, BaseRecord):
            linked_base_and_table = self.get_linked_base_and_table()
            linked_base_and_table.ensure_base_and_table_match(value)
            return _RecordLink(base_and_table=linked_base_and_table, record=value)
        return super().validate(value, base_and_table=base_and_table)

    def clone_value(self, value: Optional[_RecordLink]) -> Optional[_RecordLink]:
        if value is None:
            return None
        return _RecordLink(base_and_table=value, other=value)

    def decode_from_airtable(self, value: Optional[List[str]], base_and_table: '_BaseAndTableProtocol') \
            -> Optional[_RecordLink]:
        if not value:
            return None
        if len(value) > 1:
            raise ValueError('Multiple records returned')

        # Propagates the base_id of the referencing record
        linked_base_and_table = BaseAndTable(
            base_id=base_and_table.base_id,
            table_id=self.get_linked_base_and_table().table_id)

        return _RecordLink(base_and_table=linked_base_and_table, record_id=value[0], fetcher=self._fetcher)

    def encode_to_airtable(self, value: _RecordLink) -> Optional[List[str]]:
        if not value:
            return []
        return [value.id]


class _RecordLinkCollection(BaseAndTable, collections.abc.Collection):
    _items: List[_RecordLink]
    _fetcher = None

    def __init__(self, *,
                 base_and_table: '_BaseAndTableProtocol',
                 other: Optional[_RecordLinkCollection] = None,
                 fetcher=None):
        super(_RecordLinkCollection, self).__init__(base_id=base_and_table.base_id, table_id=base_and_table.table_id)

        self._items = []

        if other is not None:
            self.ensure_base_and_table_match(other)
            if fetcher is None:
                fetcher = other._fetcher
        self._fetcher = fetcher

        if other is not None:
            for item in other._items:
                self._items.append(_RecordLink(base_and_table=self, other=item, fetcher=self._fetcher))

    def clear(self) -> None:
        self._items.clear()

    def ids(self) -> Iterator[str]:
        return (item.id for item in self._items)

    def __bool__(self):
        return bool(self._items)

    def __len__(self) -> int:
        return len(self._items)

    def __iadd__(self, other: Union[str, 'BaseRecord']):
        from pyrtable.record import BaseRecord

        if isinstance(other, BaseRecord):
            record = other
            self.ensure_base_and_table_match(other)
            record_link = _RecordLink(
                base_and_table=self, record_id=other.id, record=other, fetcher=self._fetcher)
        else:
            record = None
            record_link = _RecordLink(base_and_table=self, record_id=other, fetcher=self._fetcher)

        try:
            existing_index = self._items.index(record_link)
            # Perhaps we have got the missing record?
            if record is not None:
                existing_record_link = self._items[existing_index]
                if not existing_record_link.has_fetched_record:
                    self._items[existing_index] = record_link

        except ValueError:
            self._items.append(record_link)

        return self

    def __isub__(self, other: Union[str, BaseRecord]):
        from pyrtable.record import BaseRecord

        if isinstance(other, BaseRecord):
            self.ensure_base_and_table_match(other)
            record_link = _RecordLink(
                base_and_table=self, record_id=other.id, record=other, fetcher=self._fetcher)
        else:
            record_link = _RecordLink(
                base_and_table=self, record_id=other, fetcher=self._fetcher)

        try:
            self._items.remove(record_link)
        except ValueError:
            pass

        return self

    def __contains__(self, value: object) -> bool:
        return value in self._items

    def __eq__(self, other):
        if not isinstance(other, _RecordLinkCollection):
            return NotImplemented
        return self._items == other._items

    def _iter_ids(self) -> Iterator[Optional[str]]:
        # noinspection PyProtectedMember
        yield from (item._id for item in self._items)

    def __iter__(self) -> Iterator[BaseRecord]:
        for item in self._items:
            yield item.record


class _MultipleRecordIdsLinkProxyField:
    def __init__(self, record_attr_name: str):
        self._record_attr_name = record_attr_name

    def __get__(self, instance, owner: 'BaseField') -> Iterator[str]:
        # noinspection PyProtectedMember
        value = instance._fields_values[self._record_attr_name]
        # noinspection PyProtectedMember
        yield from value._iter_ids()


class MultipleRecordLinkField(BaseRecordLinkField):
    @staticmethod
    def _get_ids_attr_name(attr_name: str) -> str:
        return '%s_ids' % attr_name

    @classmethod
    def _install_extra_properties(cls, record_cls: Type['BaseRecord'], attr_name: str):
        super()._install_extra_properties(record_cls, attr_name)
        setattr(record_cls, cls._get_ids_attr_name(attr_name), _MultipleRecordIdsLinkProxyField(attr_name))

    def validate(self,
                 value: Optional[Union[_RecordLinkCollection, Iterable[Any]]],
                 base_and_table: '_BaseAndTableProtocol') -> Any:
        if isinstance(value, _RecordLinkCollection):
            linked_base_and_table = self.get_linked_base_and_table()
            linked_base_and_table.ensure_base_and_table_match(value)
            return _RecordLinkCollection(base_and_table=linked_base_and_table, other=value)
        if isinstance(value, collections.abc.Iterable):
            result = _RecordLinkCollection(base_and_table=self.get_linked_base_and_table(), fetcher=self._fetcher)
            for record_id in value or []:
                result += record_id
            return result

        return super().validate(value, base_and_table=base_and_table)

    def clone_value(self, value: _RecordLinkCollection) -> _RecordLinkCollection:
        return _RecordLinkCollection(base_and_table=self.get_linked_base_and_table(), other=value)

    def decode_from_airtable(self, value: Optional[List[str]], base_and_table: '_BaseAndTableProtocol') \
            -> _RecordLinkCollection:
        result = _RecordLinkCollection(base_and_table=self.get_linked_base_and_table(), fetcher=self._fetcher)
        for record_id in value or []:
            result += record_id
        return result

    def encode_to_airtable(self, value: _RecordLinkCollection) -> Optional[List[str]]:
        if not value:
            return None
        return list(value.ids())


__all__ = ['SingleRecordLinkField', 'MultipleRecordLinkField']



import datetime
import re
from typing import TYPE_CHECKING, Type, Iterator, Optional, Dict, Any, Union, List, Callable, Tuple, Protocol

from ._baseandtable import _BaseAndTableSettableMixin
from .query import RecordQuery


try:
    import simplejson as json
except ImportError:
    import json

try:
    import pytz
except ImportError:
    pytz = None

from pyrtable.fields import BaseField


if TYPE_CHECKING:
    pass


class _MetaManager:
    table_tag: str = None
    base_tag: str = None

    def __get__(self, instance, owner):
        if self.table_tag is None:
            meta_class = getattr(owner, 'Meta')
            if meta_class is None:
                raise ValueError('No Meta class defined for class %s' % owner.__name__)

            base_tag = getattr(meta_class, 'base')
            if base_tag is None:
                raise ValueError('Meta class does not contain a "base" field')

            table_tag = getattr(meta_class, 'table')
            if table_tag is None:
                m = re.fullmatch(r'(.+)Record', owner.__name__)
                if not m:
                    raise ValueError(
                        'Meta class does not contain a "table" field,'
                        ' and the table name cannot be inferred from the class name')

                table_tag = re.sub(r'(.)([A-Z][a-z]+)', r'\1_\2', m.group(1))
                table_tag = re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', table_tag).lower()

            self.base_tag = base_tag
            self.table_tag = table_tag

        return self


class _BaseRecordProtocol(Protocol):
    """
    Protocol for :class:`BaseRecord` type hinting.
    """
    class Meta:
        api_key: str
        base_id: str
        table_id: str
        record_query_class: Type['RecordQuery']

        get_api_key: Callable[[], str]
        get_base_id: Callable[[], str]
        get_table_id: Callable[[], str]
        get_record_query_class: Callable[[], Type['RecordQuery']]


class BaseRecord(_BaseAndTableSettableMixin, _BaseRecordProtocol):
    """
    Base class for all table records.
    """
    _ATTRIBUTE_NOT_SPECIFIED = object()

    _id: Optional[str] = None
    _base_id: Optional[str] = None
    _table_id: Optional[str] = None
    _created_timestamp: Optional[datetime.datetime] = None

    meta = _MetaManager()
    objects: RecordQuery

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__()

        for attr_name, field in cls.iter_fields():
            # noinspection PyProtectedMember
            field._install_extra_properties(cls, attr_name)

        record_query_cls = cls._get_meta_attr('record_query_class', RecordQuery)
        record_query = record_query_cls(record_class=cls)
        cls.objects = record_query

    @classmethod
    def iter_fields(cls) -> Iterator[Tuple[str, BaseField]]:
        yielded_attrs = set()

        for current_cls in cls.__mro__:
            for attr_name, field in list(current_cls.__dict__.items()):
                if attr_name not in yielded_attrs and isinstance(field, BaseField):
                    yielded_attrs.add(attr_name)
                    yield attr_name, field

    @classmethod
    def get_column_names(cls) -> List[str]:
        return [field.column_name for _, field in cls.iter_fields()
                if field.column_name]

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)

        instance._fields = {}

        for attr_name, field in cls.iter_fields():
            if attr_name in ('id', 'created_timestamp'):
                raise AttributeError(f'"{attr_name}" is a reserved field name and cannot be used')
            field.attr_name = attr_name
            instance._fields[attr_name] = field
            field._record = instance

            if field.column_name is None:
                field._column_name = attr_name

        return instance

    def __init__(self, _base_id: Optional[str] = None, _table_id: Optional[str] = None, **kwargs):
        super().__init__(base_id=_base_id, table_id=_table_id, record_cls=type(self))
        self._fields_values = {}
        self._orig_fields_values = {}

        for attr_name, field in self.iter_fields():
            self._fields_values[attr_name] = field.decode_from_airtable(None, base_and_table=self)
        self._clear_dirty_fields()

        for key, value in kwargs.items():
            if key not in self._fields:
                TypeError('init() got an unexpected keyword argument %r' % key)
            setattr(self, key, value)

    def _clear_dirty_fields(self):
        for attr_name, field in self.iter_fields():
            self._orig_fields_values[attr_name] = field.clone_value(self._fields_values[attr_name])

    def consume_airtable_data(self, data: Dict[str, Any]):
        data = dict(data)
        self._id = data.pop('id')
        self._created_timestamp = datetime.datetime.strptime(data.pop('createdTime'), '%Y-%m-%dT%H:%M:%S.%fZ')
        if pytz is not None:
            self._created_timestamp = pytz.UTC.localize(self._created_timestamp)

        fields_values: Dict[str, Any] = data.pop('fields')

        for attr_name, field in self.iter_fields():
            if field.column_name is None:
                continue
            value = field.decode_from_airtable(fields_values.get(field.column_name), base_and_table=self)
            self._fields_values[attr_name] = value

        self._clear_dirty_fields()

    @property
    def id(self) -> Optional[str]:
        return self._id

    @property
    def created_timestamp(self) -> Optional[datetime.datetime]:
        return self._created_timestamp

    def delete(self) -> None:
        """
        Delete the record from Airtable.
        """
        from pyrtable.context import get_default_context
        get_default_context().delete(self.__class__, self)

    def save(self) -> None:
        """
        Save the record to Airtable.
        """
        from pyrtable.context import get_default_context
        get_default_context().save(self.__class__, self)

    def encode_to_airtable(self, include_non_dirty_fields=False) -> Dict[str, Any]:
        result = {}

        for attr_name, field in self.iter_fields():
            if field.read_only or field.column_name is None:
                continue

            value = self._fields_values[attr_name]
            if not include_non_dirty_fields \
                    and field.is_same_value(value, self._orig_fields_values[attr_name]):
                continue

            result[field.column_name] = field.encode_to_airtable(value)

        return result

    def normalize_fields(self) -> None:
        for field in self._fields.values():
            if field.normalize is not None:
                if field.skip_normalization_if_filled and getattr(self, field.attr_name):
                    # @TODO Differ False and None if the field is boolean
                    pass
                else:
                    if field.normalize_from_attr_name is not None:
                        value = getattr(self, field.normalize_from_attr_name)
                    else:
                        value = getattr(self, field.attr_name)

                    setattr(self, field.attr_name, field.normalize(value))

    def __repr__(self):
        return '<%s (%s)>' % (self.__class__.__name__, self.id)

    @classmethod
    def get_request_headers(cls, defaults=None, base_id: Optional[str] = None) -> Dict[str, str]:
        if not defaults:
            defaults = {}

        result = dict(**defaults)
        result['Authorization'] = 'Bearer %s' % cls._get_api_key(base_id=base_id)

        return result

    @classmethod
    def _get_meta_attr(cls, attr_name: str, default_value: Any = _ATTRIBUTE_NOT_SPECIFIED) -> Any:
        meta_class = getattr(cls, 'Meta', None)
        if meta_class is not None:
            if hasattr(meta_class, attr_name):
                return getattr(meta_class, attr_name)
            elif hasattr(meta_class, 'get_' + attr_name):
                return getattr(meta_class, 'get_' + attr_name)()

        for base_class in cls.__bases__:
            if issubclass(base_class, BaseRecord):
                parent_attribute_not_specified = object()
                value = base_class._get_meta_attr(attr_name=attr_name,
                                                  default_value=parent_attribute_not_specified)
                if value != parent_attribute_not_specified:
                    return value

        if default_value != BaseRecord._ATTRIBUTE_NOT_SPECIFIED:
            return default_value

        raise AttributeError("'Meta.%s' attribute is not defined for class %r" % (attr_name, cls.__name__))

    @classmethod
    def _get_api_key(cls, base_id: Optional[str] = None) -> Optional[str]:
        if hasattr(cls, 'get_api_key'):
            function = cls.get_api_key

            if base_id is not None:
                import functools
                import inspect

                signature = inspect.signature(function)
                # Can base_id be sent as a keyword argument?
                if 'base_id' in signature.parameters \
                        or any(parameter.kind is inspect.Parameter.VAR_KEYWORD
                               for parameter in signature.parameters.values()):
                    function = functools.partial(function, base_id=base_id)

            result = function()
        else:
            result = cls._get_meta_attr('api_key', default_value=None)

        return result

    @classmethod
    def get_class_base_id(cls) -> str:
        return cls._get_meta_attr('base_id', None)

    @classmethod
    def get_class_table_id(cls) -> str:
        return cls._get_meta_attr('table_id', None)


class APIKeyFromSecretsFileMixin:
    AIRTABLE_SECRETS_FILENAME = 'airtable_secrets.yaml'

    @classmethod
    def get_api_key(cls, base_id=None):
        if not issubclass(cls, BaseRecord):
            raise AttributeError('This is a mixin for BaseRecord subclasses')

        from pyrtable.configutils import load_config_file

        cls: Union[BaseRecord, APIKeyFromSecretsFileMixin]

        if base_id is None:
            base_id = cls.get_class_base_id()
        if base_id is None:
            raise ValueError('Base ID is not set')

        all_api_keys = load_config_file(cls.AIRTABLE_SECRETS_FILENAME)
        api_key = all_api_keys.get(base_id)
        if api_key is None:
            raise KeyError('API key not found in file %r: %r' % (cls.AIRTABLE_SECRETS_FILENAME, api_key))
        return api_key


class TableIDFromClassNameMixin:
    @classmethod
    def get_table_id(cls) -> str:
        if not issubclass(cls, BaseRecord):
            raise AttributeError('This is a mixin for BaseRecord subclasses')

        # Respect Meta.table_id if defined
        table_id = cls._get_meta_attr('table_id', None)
        if table_id is not None:
            return table_id

        m = re.fullmatch(r'(.+)Record', cls.__name__)
        if not m:
            raise AttributeError("Class %r does not have a name that ends with 'Record'")

        table_id = m.group(1)
        table_id = re.sub(r'([a-z])([A-Z])', r'\1 \2', table_id)
        return table_id


__all__ = ['BaseRecord', 'APIKeyFromSecretsFileMixin', 'TableIDFromClassNameMixin']
